#include <stdio.h>

#include "clwoobject.hpp"

#ifdef INCLUDE_PRINTF_DEBUG
void cLWOObject :: printLWOID(long t)
{
  printf("%c%c%c%c", ((unsigned char) (t >> 24)), ((unsigned char) (t >> 16)),
	 (unsigned char)(t >> 8), (unsigned char) t);
};
#endif


/**
 * The following methods are used to obtain double words, words or bytes from
 * the LWO file.
 */

long cLWOObject :: dwGet(FILE *fp, long &dw)
{
  fread(&dw, sizeof(long), 1, fp); 
  dw = ntohl(dw);
  return(dw);
};

short int cLWOObject :: wGet(FILE *fp, short int &w)
{
  unsigned char b;
  w = ((unsigned char)bGet(fp, b) << 8)
    + ((unsigned char)bGet(fp, b));
  return (w);
};

unsigned char cLWOObject :: bGet (FILE *fp, unsigned char &b)
{
  fread(&b, sizeof(char), 1, fp); 
  return(b);
};

long cLWOObject :: getNoItems(FILE *lwoFile)
{
  long sizePOLS, posPOLS;
  long dwChunkId, dwSize, dwType;
  bool q = false;
  
  numImages = 0, numFaces = 0;
  
  while (dwGet(lwoFile, dwChunkId) != LWO2_ERROR && !q)
    {
      dwGet(lwoFile, dwSize);
      
      switch(dwChunkId)
	{
	case LWO_ID('C','L','I','P'):
	  numImages++;
	  break;
	case LWO_ID('P','O','L','S'):
	  sizePOLS = dwSize;
	  posPOLS = ftell(lwoFile);
	  break;
	}
      
      
      if (fseek(lwoFile, dwSize, SEEK_CUR) != 0)
	q = true;
      
      if (dwSize < 0)
	q= true;
    }
  
#ifdef INCLUDE_PRINTF_DEBUG
  printf("POLS info: %li %li\n", sizePOLS, posPOLS);
#endif
  
  fseek(lwoFile, posPOLS, SEEK_SET);
  dwSize = sizePOLS;
  
  dwGet(lwoFile, dwType);
  
  if (dwType == LWO_ID('F','A','C','E'))
    {
      long dwLoop = 4; // type already read
      while (dwLoop <  dwSize)
	{
	  long dw;
	  short int wCount;
	  wGet(lwoFile, wCount);
	  wCount &= 1023;
	  numFaces++;
	  dw = wCount << 1;
	  fseek(lwoFile, dw, SEEK_CUR);
	  dwLoop += (dw+2);
	}
    }
  
  return(LWO2_OK);
};

bool cLWOObject :: processTAGS(FILE *lwoFile, long dwSize)
{
  long sPos = ftell(lwoFile) + dwSize;
  long dw, dwLoop;
  // first count number of surface "names"
  unsigned char b;
  dw = 0;
  dwLoop = 0;
  numSurfaces = 0;


  char SurfaceName[64];
  do 
    {
      bGet(lwoFile, b);
      SurfaceName[dw] = b;
      dwLoop++;
      if (b== 0)
	{
	  numSurfaces++;
	  dw=0;
	  bGet(lwoFile, b);
	  if (b==0)
	    dwLoop++; // get rid of the padding;
	  else
	    fseek(lwoFile, -1, SEEK_CUR);
	}
      else
	dw++;
    } while(dwLoop<dwSize);
  fseek(lwoFile, -dwSize, SEEK_CUR);
	
#ifdef INCLUDE_PRINTF_DEBUG	  
  printf("Number of surfaces: %li\n", numSurfaces); 
#endif
  if ((mySurfaces = new sSurface[numSurfaces]) == 0)
    {
#ifdef INCLUDE_PRINTF_DEBUG
      printf("Allocation of surfaces failed\n");
#endif
      return(false);
    }
	      
  dwLoop=0;
  long dwFillLoop=0;
  dw = 0;
  do
    {
      bGet(lwoFile, b);
      //printf("%c %i %li\n", b, dw, dwFillLoop);
		  
      mySurfaces[dwFillLoop].name[dw]=b;
      dwLoop++;
      if (b==0)
	{
#ifdef INCLUDE_PRINTF_DEBUG
	  printf("Surface %li name: %s\n", dwFillLoop, mySurfaces[dwFillLoop].name);
#endif
	  dwFillLoop++;
	  dw = 0;
	  bGet(lwoFile, b);
	  if (b==0)
	    dwLoop++;
	  else
	    fseek(lwoFile, -1, SEEK_CUR);
	}
      else
	dw++;
    } while(dwLoop<dwSize);
	      
  fseek(lwoFile, sPos , SEEK_SET); 
  return(true);
};

bool cLWOObject :: processPOLS(FILE *lwoFile, long dwSize)
{
  long dwType, dwLoop;
  bool q = false;
  dwGet(lwoFile, dwType);
	      
  switch(dwType)
    {
    case LWO_ID('F','A','C','E'):
      {
	short int wCount;
	// get NbTriangles
	numTriangles = 0;
	dwLoop=4; // 4 bytes (type)
	while(dwLoop < dwSize)
	  {	
			
	    long dw;
	    wGet(lwoFile, wCount);
	    wCount &= 1023;
	    numTriangles += (long)(wCount-2);
	    dw = wCount<<1;
	    fseek(lwoFile, dw, SEEK_CUR);
	    //IncPos(dw); // 2 bytes
	    dwLoop += (dw+2); // NbVertices + Datas
	  }
		    
#ifdef INCLUDE_PRINTF_DEBUG
	printf("Triangle count: %li\n", numTriangles);
#endif
	fseek(lwoFile, -(dwSize-4), SEEK_CUR); // 4 bytes for type
		    
	if ((myTriangles = new sTriangle[numTriangles]) == 0)
	  {
#ifdef INCLUDE_PRINTF_DEBUG
	    printf("Triangle memory allocation failed\n");
#endif
	    return(false);
	  }
		    
	short int wFace = 0, wTriangle = 0;
	while (wFace < numFaces)
	  {
	    short int wVert1, wVert2, wVert3;
	    wGet(lwoFile, wCount);
			
	    wGet(lwoFile, wVert1);
	    wGet(lwoFile, wVert2);
			
	    dwLoop = 0;
	    while(dwLoop < (long) (wCount-2))
	      {
		wGet(lwoFile, wVert3);
			    
		myTriangles[wTriangle].vert1 = wVert1;
		myTriangles[wTriangle].vert2 = wVert2;
		myTriangles[wTriangle].vert3 = wVert3;
		myTriangles[wTriangle].surfaceid = wFace;

		wVert2 = wVert3;

		wTriangle++;
		dwLoop++;
	      }
	    wFace++;
	  }
      }
		  
      break;
	      
    default:
      if(fseek(lwoFile, dwSize-4, SEEK_CUR) != 0) // type has been read
	q = true;
      break;
    }
  return(true);
};

bool cLWOObject :: processSURF(FILE *lwoFile, long dwSize)
{
  long sPos = ftell(lwoFile) + dwSize;
  short int wSurface;
  long dwLoop;

  if (numSurfaces == 0)
    {
#ifdef INCLUDE_PRINTF_DEBUG
      printf("TAGS must come before SURF\n");
#endif
      return(false);
    }

  long dwPos = 0;

  unsigned char b;
  unsigned char SurfaceName[64];
  do
    {
      bGet(lwoFile, b);
      SurfaceName[dwPos] = b;
      dwPos++;
    } while(b!=0);

#ifdef INCLUDE_PRINTF_DEBUG
  printf("%s\n", SurfaceName);
#endif

  bGet(lwoFile, b);
  dwPos++;
  while(b==0)
    {
      bGet(lwoFile, b);
      dwPos++;
    }
  fseek(lwoFile, -1, SEEK_CUR);
  dwPos--;

	      
  // now let us search for this surface
  wSurface = 0;
  while(strcmp((char*)SurfaceName, (char*)mySurfaces[wSurface].name) != 0)
    {
      wSurface++;
    }

  dwLoop=dwPos;
  long dwSubChunkId;
  while(dwLoop<dwSize) 
    {
      short int wSubSize;
      dwGet(lwoFile, dwSubChunkId);
      wGet(lwoFile, wSubSize);
		  
      dwLoop+=6;
#ifdef INCLUDE_PRINTF_DEBUG
      printf("\t");
      printLWOID(dwSubChunkId);
      printf("\n");
#endif
      switch (dwSubChunkId)
	{
	  // SMAN
		   
	case LWO_ID('B','L','O','K'):
	  {
	    short int wSubLoop = 6;
	    while (wSubLoop < wSubSize)
	      {
		long dwSubSubChunkId;
		short int wSubSubSize;
		dwGet(lwoFile, dwSubSubChunkId);
		wGet(lwoFile, wSubSubSize);
		wSubLoop += 6;
		  
		switch (dwSubSubChunkId)
		  {
		    // - PIXB
		    // - AXIS
		    // - PROJ
		    // - IMAG
		    //
		  default:
		    wSubLoop += wSubSubSize;
#ifdef INCLUDE_PRINTF_DEBUG
		    printf("\t\t");
		    printLWOID(dwSubSubChunkId);
		    printf("\n");
#endif
		    fseek(lwoFile, wSubSubSize, SEEK_CUR);
		    break;
		  }
	      }
	    dwLoop += (long)wSubSize;
	  }
	  break;

	  // SIDE

	  // ADTR

	case LWO_ID('C','O','L','R'):
	  {
	    long dwR, dwG, dwB;
	    short int w;
	    dwGet(lwoFile, dwR);
	    dwGet(lwoFile, dwG);
	    dwGet(lwoFile, dwB);
			
	    wGet(lwoFile, w);
	    // environment info
			
	    mySurfaces[wSurface].col.r = *(float*)&dwR;
	    mySurfaces[wSurface].col.g = *(float*)&dwG;
	    mySurfaces[wSurface].col.b = *(float*)&dwB;
			
	    dwLoop += (long)wSubSize;
	  }
	  break;

	default:
	  fseek(lwoFile, wSubSize, SEEK_CUR);
	  dwLoop += (long) wSubSize;
	  break;
	}
		  
    }

  fseek(lwoFile, sPos , SEEK_SET); 
  return(true);
};

void cLWOObject :: processPNTS(FILE *lwoFile, long dwSize)
{
  long sPos = ftell(lwoFile) + dwSize, dwLoop;
  dwSize = dwSize>>2; // 4 bytes
	      
#ifdef INCLUDE_PRINTF_DEBUG
  printf("Vertex count %li\n", dwSize/3);
#endif
  // allocate memory to verticies
  myVerticies = new sVertex[dwSize];
  if (myVerticies == NULL)
    {
#ifdef INCLUDE_PRINTF_DEBUG
      printf("Memory allocation failed\n");
#endif
      return;
    }	   
	      
  numVerticies = dwSize / 3;
  dwLoop = 0;
  while (dwLoop < numVerticies)
    {
      long myX, myY, myZ;
      dwGet(lwoFile, myX);
      dwGet(lwoFile, myY);
      dwGet(lwoFile, myZ);

#define sizecoef 1.0
		  
      myVerticies[dwLoop].x = sizecoef * (*(float*)&myX);
      myVerticies[dwLoop].y = sizecoef * (*(float*)&myY);
      myVerticies[dwLoop].z = sizecoef * (*(float*)&myZ);		  
      dwLoop++;
    }
	      
  fseek(lwoFile, sPos , SEEK_SET); 
};

bool cLWOObject :: load(char *filename)
{
  FILE *lwoFile;
  long dw, dwFileSize; // double word - 32bit

  if ((lwoFile = fopen(filename, "rb")) == NULL)
    {
      bExists = false;
      return(false);
    }

  dwGet(lwoFile, dw); // FORM
  dwGet(lwoFile, dwFileSize); // SIZE
  dwGet(lwoFile, dw);
#ifdef INCLUDE_PRINTF_DEBUG
  if (dw != LWO_ID('L','W','O','2'))
    printf ("This is not a valid LWO2 file\n");
  else
    printf("This is a valid LWO2 file\n");
#endif

  long cpos = ftell(lwoFile);

  //dwGet(lwoFile, dw);
    
  getNoItems(lwoFile);
#ifdef INCLUDE_PRINTF_DEBUG
  printf("Number Images: %li Number Faces: %li \n", numImages, numFaces); 
#endif
  // hello
  fseek(lwoFile, cpos, SEEK_SET);
    
  // allocate memory to images and faces
  // ignore images atm
  myFaces = new sFace[numFaces];
  if (myFaces == NULL)
    {
#ifdef INCLUDE_PRINTF_DEBUG
      printf("Face Memory allocation failed\n");
#endif
      return(false);
    }
    
  bool q = false;
  long dwChunkId, dwSize, dwLoop;
    
  while (dwGet(lwoFile, dwChunkId) != LWO2_ERROR && !q)
    {
      dwGet(lwoFile, dwSize);
#ifdef INCLUDE_PRINTF_DEBUG
      printLWOID(dwChunkId);
      printf(": size, pos: %lx %lx\n", dwSize, ftell(lwoFile));
#endif

      switch(dwChunkId)
	{
	case LWO_ID('P','N','T','S'):
	  processPNTS(lwoFile, dwSize);
	  break;

	case LWO_ID('B','B','O','X'):
	  {
#ifdef INCLUDE_PRINTF_DEBUG
	    printf("found BBOX structure\n");
#endif
	    long dwVx, dwVy, dwVz;
	    dwGet(lwoFile, dwVx);
	    bbo.min.x = *(float*)&dwVx;
	    dwGet(lwoFile, dwVy);
	    bbo.min.y = *(float*)&dwVy;
	    dwGet(lwoFile, dwVz);
	    bbo.min.z = *(float*)&dwVz;
	      
	    dwGet(lwoFile, dwVx);
	    bbo.max.x = *(float*)&dwVx;
	    dwGet(lwoFile, dwVy);
	    bbo.max.y = *(float*)&dwVy;
	    dwGet(lwoFile, dwVz);
	    bbo.max.z = *(float*)&dwVz;
	  }
	  break;

	case LWO_ID('P','O','L','S'):
	  if (!processPOLS(lwoFile, dwSize))
	    return false;
	  break;

	case LWO_ID('T','A','G','S'):
	  {
	    if (!processTAGS(lwoFile, dwSize))
	      return false;
	  }
	  break;
	    

	  // CLIP
	case LWO_ID('C','L','I','P'):
	  {
	    long dwType;
	    short int wImage;
	    if (numSurfaces == 0)
	      {
#ifdef INCLUDE_PRINTF_DEBUG
		printf("No surfaces to work on\n");
#endif
		return false;
	      }

	    dwGet(lwoFile, dw); // image index
	    wImage = (short int) dw-1; // 0 based index
	      
	    dwGet(lwoFile, dwType);
	      
	    switch(dwType)
	      {
	      case LWO_ID('S','T','I','L'):
		{
		  short int w;
		  wGet(lwoFile, w);
		    
		  char texturename[64];
		  fread(texturename, sizeof(char), w, lwoFile);
#ifdef INCLUDE_PRINTF_DEBUG
		  printf("%s\n", texturename);
#endif
		}
		break;
	      default:
		fseek(lwoFile, dwSize-8, SEEK_CUR);
		break;
	      }
	  }
	  break;

	  // SURF
	case LWO_ID('S','U','R','F'):
	  {
	    if (!processSURF(lwoFile, dwSize))
	      return false;
	  }
	  break;
	    
	  // PTAG
	case LWO_ID('P','T','A','G'):
	  {
	    long sPos = ftell(lwoFile) + dwSize;
	    long dwType;
	    dwGet(lwoFile, dwType);
	     
#ifdef INCLUDE_PRINTF_DEBUG
	    printf("Processing PTAGS\n");
#endif
	    switch (dwType)
	      {
	      case LWO_ID('S','U','R','F'):
		dwLoop = 4; // 4 bytes = type
		while(dwLoop < dwSize)
		  {
		    short int wFace, wSurface;
		    wGet(lwoFile, wFace);
		    wGet(lwoFile, wSurface);

		    //printf("%li %li\n", wFace, wSurface);
		      
		    myFaces[wFace].mySurface = wSurface;
		    dwLoop += 4;
		  }

		break;

	      default:
		fseek(lwoFile, dwSize-4, SEEK_CUR);
		break;
	      }
	    fseek(lwoFile, sPos , SEEK_SET); 
	  }
	  break;
	    
	default:
	  if (fseek(lwoFile, dwSize, SEEK_CUR) != 0)
	    q = true;

	  if (dwSize == 0)
	    q = true;
	}
      if (ftell(lwoFile) >= dwFileSize + 8)
	q = true;

    }
    
  return true;
};

/**
 * The following methods are public methods to retrieve enough information 
 * to draw the object loaded.
 *
 */

cLWOObject :: cLWOObject(char *filename)
{
  bExists = false;
  if (load(filename))
    {
      bExists = true;
#ifdef INCLUDE_PRINTF_DEBUG
      printf("loaded successfully\n");
#endif
    }
#ifdef INCLUDE_PRINTF_DEBUG
  else
    printf("load failed\n");
#endif
};

cLWOObject :: ~cLWOObject()
{
  if (bExists)
    {
      delete myVerticies;
      delete myFaces;
    }
};

long cLWOObject :: noVerticies() const
{
  return(numVerticies);
};

sVertex cLWOObject :: getVertex(long i) const
{
#ifdef INCLUDE_PRINTF_DEBUG
  if (i > numVerticies)
    printf("%liwtf\n", i);
#endif
  return(myVerticies[i]);
};

long cLWOObject :: noTriangles() const
{
  return(numTriangles);
};

sTriangle cLWOObject :: getTriangle(long i) const
{
  return(myTriangles[i]);
};

sSurface cLWOObject :: getSurface(long i) const
{
  return(mySurfaces[myFaces[i].mySurface]);
};
