#ifndef __CLWOOBJECT_HPP
#define __CLWOOBJECT_HPP

#include <string.h>
#include <netinet/in.h>

#define INCLUDE_PRINTF_DEBUG

/** Object to load an LWO Object from a LWO2 file
 *  and store it in a data-structure suitable for this.
 *
 * maybe later will also add plugins for LWOB files too
 * as well as a drawing function and other things.
 */


struct sVertex
{ 
  float x, y, z;
};

struct bbox
{
  sVertex min, max;
};

struct sTriangle
{
  short int vert1, vert2, vert3;
  long surfaceid;
};

struct sFace
{
  
  long mySurface;
};

struct sColor
{
  float r, b, g;
};

struct sSurface
{
  char name[64];
  sColor col;
};



#define LWO_ID(d,c,b,a)     (((d)<<24) | ((c)<<16) | ((b)<<8) | (a))

enum LWO2RESULT
{
	LWO2_OK = 0,
	LWO2_ERROR = 1
};

class cLWOObject {

private: 
  bool bExists;
  long numImages, numFaces, numVerticies, numTriangles, numSurfaces;
  bbox bbo;
  sFace *myFaces;
  sVertex *myVerticies;
  sTriangle *myTriangles;
  sSurface *mySurfaces;

#ifdef INCLUDE_PRINTF_DEBUG
  // debug method.
  void printLWOID(long t);
#endif
  
  // information getters.
  static long dwGet(FILE *fp, long &dw);
  static short int wGet(FILE *fp, short int &w);
  static unsigned char bGet(FILE *fp, unsigned char &b);

  // processing methods.
  long getNoItems(FILE *lwoFile);
  bool processTAGS(FILE *lwoFile, long dwSize);
  bool processPOLS(FILE *lwoFile, long dwSize);
  bool processSURF(FILE *lwoFile, long dwSize);
  void processPNTS(FILE *lwoFile, long dwSize);
  bool load(char *filename);

public:
  cLWOObject(char *filename);
  ~cLWOObject();

  long noVerticies() const;
  sVertex getVertex(long i) const;

  long noTriangles() const;
  sTriangle getTriangle(long i) const;

  sSurface getSurface(long i) const; 

};

#endif
