#include <allegro.h>

/* Note: Under Windows, if you include any Windows headers (e.g. GL/gl.h),
 * you must include <winalleg.h> first.  To save you working out when this
 * occurs, `alleggl.h' handles this for you, and includes <GL/gl.h>.  If 
 * you need other GL headers, include them afterwards.
 *
 * Also note that user programs, outside of this library distribution,
 * should use <...> notation for `alleggl.h'.
 */
#include "alleggl.h"
#include <GL/glu.h>
#include <stdio.h>

#include <string.h>

#include "clwoobject.hpp"

void display (void);


struct camera {
	double xangle, yangle, zangle;
	double dist;
};

struct camera camera = {
	0.0, 0.5, 0.0,
	20.0
};

double angle_speed = 5.0;
double dist_speed = 1.0;

int frames = 0;
volatile int secs;

bool lp;
bool bp;
bool fp;

bool light;
bool blend;	// Blending OFF/ON? ( NEW )

GLfloat z=-5.0f; // depth into the screen.

GLfloat xrot;   // x rotation 
GLfloat yrot;   // y rotation 
GLfloat xspeed; // x rotation speed
GLfloat yspeed; // y rotation speed

/* white ambient light at half intensity (rgba) */
GLfloat LightAmbient[] = { 0.2f, 0.2f, 0.2f, 1.0f };

/* super bright, full intensity diffuse light. */
GLfloat LightDiffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };

/* position of light (x, y, z, (position of light)) */
GLfloat LightPosition[] = { 0.0f, 0.0f, 2.0f, 1.0f };

GLuint	filter;			/* Which Filter To Use (nearest/linear/mipmapped) */
GLuint	texture[3];		/* Storage for 3 textures. */

cLWOObject *object;

void secs_timer(void)
{
	secs++;
}
END_OF_FUNCTION(secs_timer);


GLvoid LoadGLTextures(GLvoid) {	

  // Load Texture
  BITMAP *image1;
  PALETTE pal;

  // allocate space for texture
  image1 = load_bitmap("crate.bmp", pal);       

  // Create Textures	
  glGenTextures(3, &texture[0]);

  // texture 1 (poor quality scaling)
  glBindTexture(GL_TEXTURE_2D, texture[0]);   // 2d texture (x and y size)

  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST); // cheap scaling when image bigger than texture
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); // cheap scaling when image smalled than texture

  // 2d texture, level of detail 0 (normal), 3 components (red, green, blue), x size from image, y size from image, 
    // border 0 (normal), rgb color data, unsigned byte data, and finally the data itself.
  glTexImage2D(GL_TEXTURE_2D, 0, allegro_gl_get_texture_format(image1), 
	       image1->w, image1->h, 0,  
	       (GLenum)allegro_gl_bitmap_color_format(image1), 
	       (GLenum)allegro_gl_bitmap_type(image1), image1->line[0]);

  // texture 2 (linear scaling)
  glBindTexture(GL_TEXTURE_2D, texture[1]);   // 2d texture (x and y size)
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); // scale linearly when image bigger than texture
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); // scale linearly when image smalled than texture
  glTexImage2D(GL_TEXTURE_2D, 0, allegro_gl_get_texture_format(image1), 
	       image1->w, image1->h, 0,  
	       (GLenum)allegro_gl_bitmap_color_format(image1), 
	       (GLenum)allegro_gl_bitmap_type(image1), image1->line[0]);

  // texture 3 (mipmapped scaling)
  glBindTexture(GL_TEXTURE_2D, texture[2]);   // 2d texture (x and y size)
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); // scale linearly when image bigger than texture
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR); // scale linearly + mipmap when image smalled than texture

  /* Because crummy NVidia drivers can't take GL_CLAMP *and* mipmapping at the same time... */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  /*
  glTexImage2D(GL_TEXTURE_2D, 0, allegro_gl_get_texture_format(image1), 
	       image1->w, image1->h, 0,  
	       (GLenum)allegro_gl_bitmap_color_format(image1), 
	       (GLenum)allegro_gl_bitmap_type(image1), image1->line[0]);
  */
  // 2d texture, 3 colors, width, height, RGB in that order, byte data, and the data.
  gluBuild2DMipmaps(GL_TEXTURE_2D, allegro_gl_get_texture_format(image1),
		    image1->w, image1->h, 
		    allegro_gl_bitmap_color_format(image1),
		    allegro_gl_bitmap_type(image1), image1->line[0]);
};


void set_camera_position (void)
{
	allegro_gl_begin();
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	glFrustum (-1.0, 1.0, -1.0, 1.0, 1.0, 40.0);
	glTranslatef (0, 0, -camera.dist);
	glRotatef (camera.xangle, 1, 0, 0);
	glRotatef (camera.yangle, 0, 1, 0);
	glRotatef (camera.zangle, 0, 0, 1);
	glMatrixMode (GL_MODELVIEW);
	allegro_gl_end();
}

void keyboard (void)
{
  usleep(100);
  if (key[KEY_LEFT]) camera.yangle += angle_speed;
  if (key[KEY_RIGHT]) camera.yangle -= angle_speed;
  
  if (key[KEY_UP]) camera.xangle += angle_speed;
  if (key[KEY_DOWN]) camera.xangle -= angle_speed;
  
  if (key[KEY_PGUP]) camera.dist -= dist_speed;
  if (key[KEY_PGDN]) camera.dist += dist_speed;
  
  if (key[KEY_L] && !lp) {
    lp = true;
    
    light = light ? 0 : 1;
    if (!light) {
      glDisable(GL_LIGHTING);
      printf("light off\n");
    } else {
      glEnable(GL_LIGHTING);
      printf("light on\n");
    }
    
  } else if (!key[KEY_L]) {
    lp = false;
  }

  if (key[KEY_F] && !fp) {
    fp = true;
      
    filter ++;
    printf("filter is: %i\n", filter);
    if (filter > 2) {
      filter = 0;  
    }
  } else if (!key[KEY_F]) {
    fp = false;
  }

  if (key[KEY_B] & !bp) {
    bp = true;
    
    blend = !blend;
    if (blend)
      {
	printf("blend on\n");
	glEnable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
      }
    else
      {
	printf("blend off\n");
	glDisable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
      }
  } else if (!key[KEY_B]) {
    bp = false;
  }

  if (key[KEY_4_PAD]) yspeed += 0.02f;
  if (key[KEY_6_PAD]) yspeed -= 0.02f;

  if (key[KEY_8_PAD]) xspeed += 0.02f;
  if (key[KEY_2_PAD]) xspeed -= 0.02f;


  set_camera_position();
  display();
}

GLvoid glLoadObject(char *file)
{
  object = new cLWOObject(file);
  sleep(1);
}

#define OPENGL

GLvoid glDrawObject()
{
  glColor4f(1.0, 1.0, 1.0, 0.5);
  glBegin(GL_TRIANGLES); // object->noTriangles()
  for (long k=0; k<object->noTriangles(); k++)
    {
      //printf("%li/%li ", k, object->noTriangles());
      sTriangle t = object->getTriangle(k);
      //printf("%li,%li,%li ", t.vert1, t.vert2, t.vert3);
      sVertex v1 = object->getVertex(t.vert1);
      sVertex v2 = object->getVertex(t.vert2);
      sVertex v3 = object->getVertex(t.vert3);
      
      sSurface s = object->getSurface(t.surfaceid);
      sColor c = s.col;
      glColor4f(c.r, c.g, c.b, 0.5);
      //printf("%f,%f,%f\n", c.r, c.g, c.b);
      
      glTexCoord2d(1,1); glVertex3f(v1.x,v1.y,v1.z); // Top Right
      glTexCoord2d(0,1); glVertex3f(v2.x,v2.y,v2.z); // Top Left
      glTexCoord2d(1,0); glVertex3f(v3.x,v3.y,v3.z); // Bottom Right
      //glColor4f(1.0, 0.0, 0.0, 0.5);
    }
  glEnd();
}

GLvoid showTriangles()
{
    for (long k=0; k<object->noTriangles(); k++)
    {
      //printf("%li/%li ", k, object->noTriangles());
      sTriangle t = object->getTriangle(k);
      //printf("%li,%li,%li ", t.vert1, t.vert2, t.vert3);
      sVertex v1 = object->getVertex(t.vert1);
      sVertex v2 = object->getVertex(t.vert2);
      sVertex v3 = object->getVertex(t.vert3);
      
      printf("%li. %i %i %i [%.2f %.2f %.2f] [%.2f %.2f %.2f] [%.2f %.2f %.2f] (%li) \n", k, t.vert1, t.vert2, t.vert3, v1.x, v1.y, v1.z, v2.x, v2.y, v2.z, v3.x, v3.y, v3.z, t.surfaceid ); 
    }
}

GLvoid glDrawCube()					// Draw A Cube
{
  glBegin(GL_QUADS);			// Start Drawing Quads
  for (double k=0.7; k<1.3; k+=0.1)
    {
      // Front Face
      glNormal3f( 0.0f, 0.0f, 1.0f);		// Normal Facing Forward
      glTexCoord2f(0.0f, 0.0f); glVertex3f(-k, -k,  k);	// Bottom Left Of The Texture and Quad
      glTexCoord2f(1.0f, 0.0f); glVertex3f( k, -k,  k);	// Bottom Right Of The Texture and Quad
      glTexCoord2f(1.0f, 1.0f); glVertex3f( k,  k,  k);	// Top Right Of The Texture and Quad
      glTexCoord2f(0.0f, 1.0f); glVertex3f(-k,  k,  k);	// Top Left Of The Texture and Quad
      // Back Face
      glNormal3f( 0.0f, 0.0f,-1.0f);		// Normal Facing Away
      glTexCoord2f(1.0f, 0.0f); glVertex3f(-k, -k, -k);	// Bottom Right Of The Texture and Quad
      glTexCoord2f(1.0f, 1.0f); glVertex3f(-k,  k, -k);	// Top Right Of The Texture and Quad
      glTexCoord2f(0.0f, 1.0f); glVertex3f( k,  k, -k);	// Top Left Of The Texture and Quad
      glTexCoord2f(0.0f, 0.0f); glVertex3f( k, -k, -k);	// Bottom Left Of The Texture and Quad
      // Top Face
      glNormal3f( 0.0f, 1.0f, 0.0f);		// Normal Facing Up
      glTexCoord2f(0.0f, 1.0f); glVertex3f(-k,  k, -k);	// Top Left Of The Texture and Quad
      glTexCoord2f(0.0f, 0.0f); glVertex3f(-k,  k,  k);	// Bottom Left Of The Texture and Quad
      glTexCoord2f(1.0f, 0.0f); glVertex3f( k,  k,  k);	// Bottom Right Of The Texture and Quad
      glTexCoord2f(1.0f, 1.0f); glVertex3f( k,  k, -k);	// Top Right Of The Texture and Quad
      // Bottom Face
      glNormal3f( 0.0f,-1.0f, 0.0f);		// Normal Facing Down
      glTexCoord2f(1.0f, 1.0f); glVertex3f(-k, -k, -k);	// Top Right Of The Texture and Quad
      glTexCoord2f(0.0f, 1.0f); glVertex3f( k, -k, -k);	// Top Left Of The Texture and Quad
      glTexCoord2f(0.0f, 0.0f); glVertex3f( k, -k,  k);	// Bottom Left Of The Texture and Quad
      glTexCoord2f(1.0f, 0.0f); glVertex3f(-k, -k,  k);	// Bottom Right Of The Texture and Quad
      // Right face
      glNormal3f( 1.0f, 0.0f, 0.0f);		// Normal Facing Right
      glTexCoord2f(1.0f, 0.0f); glVertex3f( k, -k, -k);	// Bottom Right Of The Texture and Quad
      glTexCoord2f(1.0f, 1.0f); glVertex3f( k,  k, -k);	// Top Right Of The Texture and Quad
      glTexCoord2f(0.0f, 1.0f); glVertex3f( k,  k,  k);	// Top Left Of The Texture and Quad
      glTexCoord2f(0.0f, 0.0f); glVertex3f( k, -k,  k);	// Bottom Left Of The Texture and Quad
      // Left Face
      glNormal3f(-1.0f, 0.0f, 0.0f);		// Normal Facing Left
      glTexCoord2f(0.0f, 0.0f); glVertex3f(-k, -k, -k);	// Bottom Left Of The Texture and Quad
      glTexCoord2f(1.0f, 0.0f); glVertex3f(-k, -k,  k);	// Bottom Right Of The Texture and Quad
      glTexCoord2f(1.0f, 1.0f); glVertex3f(-k,  k,  k);	// Top Right Of The Texture and Quad
      glTexCoord2f(0.0f, 1.0f); glVertex3f(-k,  k, -k);	// Top Left Of The Texture and Quad
    }
  glEnd();					// Done Drawing Quads
}


/* The main drawing function. */

void display()
{
  allegro_gl_begin();

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);		// Clear The Screen And The Depth Buffer
    glLoadIdentity();				// Reset The View

    glTranslatef(0.0f,0.0f,z);                  // move z units out from the screen.
    glTranslatef(0.0f,-4,0);
    
    glRotatef(xrot,1.0f,0.0f,0.0f);		// Rotate On The X Axis
    glRotatef(yrot,0.0f,1.0f,0.0f);		// Rotate On The Y Axis

    glBindTexture(GL_TEXTURE_2D, texture[filter]);   // choose the texture to use.

    glDrawObject();

    
    xrot+=xspeed;		// Increase Rotation On X Axis
    yrot+=yspeed;		// Increase Rotation On Y Axis


  glFlush();

  // swap buffers to display, since we're double buffered.
  allegro_gl_flip();
  allegro_gl_end();

  frames++;
  //glutSwapBuffers();
}


void InitGL()
{
    LoadGLTextures();
    glEnable(GL_TEXTURE_2D);
    //glClearColor(0.5f,0.5f,0.5f,1.0f);
    glClearColor (0, 0, 0, 0);
    glClearDepth(1.0);
    glDepthFunc(GL_LESS);

    glShadeModel (GL_SMOOTH); // instead of GL_FLAT
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
    glPolygonMode (GL_BACK, GL_POINTS);
    glEnable (GL_DEPTH_TEST);
    glCullFace (GL_BACK);
    glEnable (GL_CULL_FACE);

    glBlendFunc(GL_SRC_ALPHA,GL_ONE);		// Blending Function For Translucency Based On Source Alpha Value ( NEW )
    
    glMatrixMode(GL_MODELVIEW);
    
    // set up light number 1.
    glLightfv(GL_LIGHT1, GL_AMBIENT, LightAmbient);  // add lighting. (ambient)
    glLightfv(GL_LIGHT1, GL_DIFFUSE, LightDiffuse);  // add lighting. (diffuse).
    glLightfv(GL_LIGHT1, GL_POSITION,LightPosition); // set light position.
    glEnable(GL_LIGHT1);                             // turn light 1 on.  
    

    glColor4f(1.0f,1.0f,1.0f,0.5f);
}


int main (int argc, char *argv[])
{
  allegro_init();

#ifdef OPENGL  
  install_allegro_gl();
  
  allegro_gl_clear_settings();
  allegro_gl_set (AGL_COLOR_DEPTH, 16);
  allegro_gl_set (AGL_Z_DEPTH, 16);
  allegro_gl_set (AGL_DOUBLEBUFFER, 1);
  allegro_gl_set (AGL_SUGGEST, AGL_COLOR_DEPTH | AGL_Z_DEPTH | AGL_DOUBLEBUFFER);
  
  set_color_depth (16);
  if (set_gfx_mode(GFX_OPENGL, 640, 480, 0, 0) < 0) {
    allegro_message ("Error setting graphics mode:\n%s\n", allegro_error);
    return -1;
  }
  
  install_keyboard();
  install_timer();
  
  LOCK_FUNCTION(secs_timer);
  LOCK_VARIABLE(secs);
  
  allegro_gl_begin();
  
  InitGL();
  allegro_gl_end();
  
  
  
  install_int (secs_timer, 1000);
#endif
  /* lets see if we can run */
  if (argc > 1) {
    glLoadObject(argv[1]);
#ifdef OPENGL
  showTriangles();
  do {
    keyboard();
  } while (!key[KEY_ESC]);
  
  } else {
    allegro_message("Please supply file to view on command line\n");
  }
#ifndef _WIN32
  set_gfx_mode (GFX_TEXT, 0, 0, 0, 0);
#endif
  
  allegro_message("Frames: %i, Seconds: %i, FPS: %f\n\n", frames, secs, (float)frames / (float)secs);
#endif
  return 0;
}
END_OF_MAIN();
